using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

namespace ScsMinimapper.ImageFormats {
	/// <summary>
	/// Class for loading (uncompressed) DDS images (buildings and roads)
	/// </summary>
	class DdsFile {
		private string filename;
		private string errorMessage;
		private const string stderr = "The buildings image could not be loaded: ";

		/// <summary>
		/// Error message (set if GetImage() returns null)
		/// </summary>
		public string ErrorMessage {
			get { return errorMessage; }
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="filename">File to load</param>
		public DdsFile(string filename) {
			this.filename = filename;
		}

		/// <summary>
		/// Loads the image from the DDS file and returns it
		/// </summary>
		/// <returns>Image of the DDS file. Null on error</returns>
		public Bitmap GetImage() {
			FileStream fs = null;
			Bitmap image = null;
			try {
				fs = new FileStream(filename, FileMode.Open, FileAccess.Read);
				image = ReadImage(fs);
			} catch (FormatException) {
				errorMessage = stderr + "File is not a DDS image";
			} catch (NotSupportedException) {
				errorMessage = stderr + "Unsupported DDS format";
			} catch (EndOfStreamException) {
				errorMessage = stderr + "Unexpected end of file";
			} catch (FileNotFoundException) {
				errorMessage = stderr + "File not found";
			} catch (IOException) {
				errorMessage = stderr + "Error reading file";
			} catch (Exception ex) {
				errorMessage = stderr + ex.Message;
			} finally {
				if (fs != null) fs.Close();
			}
			return image;
		}

		/// <summary>
		/// Reads the image from the file stream
		/// </summary>
		/// <param name="fs">File stream to read from</param>
		/// <returns>Image</returns>
		private Bitmap ReadImage(FileStream fs) {
			// DDS header
			fs.Seek(4, SeekOrigin.Begin); // Skip "DDS" fourcc
			if (ReadInt(fs) != 124) {
				// Not a DDS image
				throw new FormatException();
			}
			fs.Seek(4, SeekOrigin.Current);
			int width = ReadInt(fs);
			int height = ReadInt(fs);

			fs.Seek(84, SeekOrigin.Begin);
			int type = ReadInt(fs);
			if (type != 0) {
				// Not an uncompressed DDS image: unsupported
				throw new NotSupportedException();
			}
			fs.Seek(124, SeekOrigin.Begin);

			// The actual pixel data
			byte[] pixels = new byte[width * height * 4];
			int offset = 0;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					pixels[offset++] = (byte)fs.ReadByte(); // b
					pixels[offset++] = (byte)fs.ReadByte(); // g
					pixels[offset++] = (byte)fs.ReadByte(); // r
					pixels[offset++] = (byte)fs.ReadByte(); // a
				}
			}

			// Assemble the actual bitmap
			Bitmap image = new Bitmap(width, height, PixelFormat.Format32bppArgb);
			Rectangle rect = new Rectangle(0, 0, width, height);
			BitmapData pixelData = image.LockBits(rect, ImageLockMode.WriteOnly, image.PixelFormat);
			System.Runtime.InteropServices.Marshal.Copy(pixels, 0, pixelData.Scan0, 4 * width * height);
			image.UnlockBits(pixelData);

			return image;

		}

		/// <summary>
		/// Reads an int from the stream
		/// </summary>
		/// <param name="fs">Stream to read from</param>
		/// <returns>Integer that has just been read</returns>
		private int ReadInt(FileStream fs) {
			byte[] data = new byte[4];
			fs.Read(data, 0, 4);

			int result = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
			return result;
		}
	}
}
