using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

namespace ScsMinimapper.ImageFormats {
	/// <summary>
	/// Class for loading (uncompressed) TGA images (terrain)
	/// </summary>
	class TgaFile {
		private string filename;
		private string errorMessage;
		private const string stderr = "The terrain image could not be loaded: ";

		/// <summary>
		/// Error message (set if GetImage() returns null)
		/// </summary>
		public string ErrorMessage {
			get { return errorMessage; }
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="filename">File to load</param>
		public TgaFile(string filename) {
			this.filename = filename;
		}

		/// <summary>
		/// Loads the image from the TGA file and returns it
		/// </summary>
		/// <returns>Image of the TGA file. Null on error</returns>
		public Bitmap GetImage() {
			FileStream fs = null;
			Bitmap image = null;
			try {
				fs = new FileStream(filename, FileMode.Open, FileAccess.Read);
				image = ReadImage(fs);
			} catch (NotSupportedException) {
				errorMessage = stderr + "Unsupported TGA format";
			} catch (EndOfStreamException) {
				errorMessage = stderr + "Unexpected end of file";
			} catch (FileNotFoundException) {
				errorMessage = stderr + "File not found";
			} catch (IOException) {
				errorMessage = stderr + "Error reading file";
			} catch (Exception ex) {
				errorMessage = stderr + ex.Message;
			} finally {
				if (fs != null) fs.Close();
			}
			return image;
		}

		/// <summary>
		/// Reads the image from the file stream
		/// </summary>
		/// <param name="fs">File stream to read from</param>
		/// <returns>Image</returns>
		private Bitmap ReadImage(FileStream fs) {
			// TGA header
			int comment = fs.ReadByte();
			fs.ReadByte();
			if (fs.ReadByte() != 2) {
				// Unsupported TGA format
				throw new NotSupportedException();
			}
			fs.Seek(12, SeekOrigin.Begin);
			int width = fs.ReadByte() | fs.ReadByte() << 8;
			int height = fs.ReadByte() | fs.ReadByte() << 8;
			fs.Seek(18 + comment, SeekOrigin.Begin);

			// The actual pixel data
			byte[] pixels = new byte[width * height * 3];
			int offset = 0;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					pixels[offset++] = (byte)fs.ReadByte(); // b
					pixels[offset++] = (byte)fs.ReadByte(); // g
					pixels[offset++] = (byte)fs.ReadByte(); // r
				}
			}

			// Assemble the actual bitmap
			Bitmap image = new Bitmap(width, height, PixelFormat.Format24bppRgb);
			Rectangle rect = new Rectangle(0, 0, width, height);
			BitmapData pixelData = image.LockBits(rect, ImageLockMode.WriteOnly, image.PixelFormat);
			System.Runtime.InteropServices.Marshal.Copy(pixels, 0, pixelData.Scan0, 3 * width * height);
			image.UnlockBits(pixelData);

			return image;
		}
	}
}
