using System;

namespace ScsMinimapper.Scs {
	/// <summary>
	/// Exception for problems with the input
	/// </summary>
	class InputException : Exception {

		private int position;

		/// <summary>
		/// Position in the stream at which the error occurred
		/// </summary>
		public int Position {
			get { return position; }
		}

		/// <summary>
		/// Creates a new input exception
		/// </summary>
		/// <param name="msg">Error message</param>
		/// <param name="pos">Position of the error in the stream</param>
		public InputException(string msg, int pos)
			: base(msg) {
			position = pos;
		}
	}

	/// <summary>
	/// Class for reading data from a memory buffer
	/// </summary>
	class Input {

		private byte[] data;
		private int ptr;
		private int length;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="data">Data buffer</param>
		public Input(byte[] data) {
			this.data = data;
			ptr = 0;
			length = data.Length;
		}

		/// <summary>
		/// Reads a single byte from the buffer
		/// </summary>
		/// <returns>The next byte in the stream</returns>
		public uint ReadByte() {
			if (ptr >= length) { // safety check
				throw new InputException(
					"Attempting to read past end of file", ptr);
			}

			return (uint)data[ptr++];
		}

		/// <summary>
		/// Reads an unsigned integer (little endian, 4 bytes) from the buffer
		/// </summary>
		/// <returns>The unsigned integer encoded at the next four bytes
		/// of the stream</returns>
		public uint ReadInt() {
			uint result;

			if (ptr > length - 4) { // safety check
				throw new InputException(
					"Attempting to read past end of file", ptr);
			}

			result = (uint)(data[ptr + 3] << 24)
				   | (uint)(data[ptr + 2] << 16)
				   | (uint)(data[ptr + 1] << 8)
				   | (uint)data[ptr];

			ptr += 4;
			return result;
		}

		/// <summary>
		/// Sets the input pointer to the specified position. This is
		/// equivalent to Seek() on streams with the SeekOrigin parameter
		/// set to SeekOrigin.Begin
		/// </summary>
		/// <param name="position">The position to set the input
		/// pointer to</param>
		public void Seek(int position) {
			ptr = position;
		}

		/// <summary>
		/// Returns the current position of the input pointer in the stream.
		/// </summary>
		/// <returns>Current position</returns>
		public int Pos() {
			return ptr;
		}

		/// <summary>
		/// Returns the total length of the stream
		/// </summary>
		/// <returns>Length of the stream</returns>
		public int Length() {
			return length;
		}

		/// <summary>
		/// Skips the specified amount of bytes in the input stream. This
		/// is equivalent to Seek() on streams with SeekOrigin set to
		/// SeekOrigin.Current
		/// </summary>
		/// <param name="amount">Number of bytes to skip</param>
		public void SkipBytes(int amount) {
			ptr += amount;
		}

		/// <summary>
		/// Skips a (wide char) input string. Such a string is stored as an
		/// integer representing the length of the string, and then the
		/// string itself with two bytes per character
		/// </summary>
		/// <exception cref="InputException">If the string length is greater
		/// than 1000 (possibly corrupt input)</exception>
		public void SkipString() {
			uint size;

			size = ReadInt();
			if (size > 1000) {
				throw new InputException("String too long", ptr);
			}
			ptr += 2 * (int)size;
		}

		/// <summary>
		/// Finds a pattern in the input. Starts searching from the current
		/// position and continues until the end of the stream.
		/// If the pattern is found, the input pointer will be at the first
		/// byte after the pattern. If the pattern is not found, the input
		/// pointer stays at the same position.
		/// </summary>
		/// <param name="pattern">Pattern to search for</param>
		/// <returns>True if the pattern is found, false otherwise</returns>
		public bool FindPattern(byte[] pattern) {
			return FindPattern(pattern, length);
		}

		/// <summary>
		/// Finds a pattern in the input. Starts searching from the current
		/// position and continues until 'maxPosition'.
		/// If the pattern is found, the input pointer will be at the first
		/// byte after the pattern. If the pattern is not found, the input
		/// pointer stays at the same position.
		/// </summary>
		/// <param name="pattern">Pattern to search for</param>
		/// <param name="maxPosition">Maximum file position</param>
		/// <returns>True if the pattern is found, false otherwise</returns>
		public bool FindPattern(byte[] pattern, int maxPosition) {
			if (ptr >= maxPosition || pattern.Length == 0 || pattern.Length > 256) {
				return false;
			}
			int[] bad_char_skip = new int[256];
			int hlen = maxPosition - ptr;
			int nlen = pattern.Length;

			for (int i = 0; i < 256; i++) {
				bad_char_skip[i] = nlen;
			}

			int last = nlen - 1;
			for (int i = 0; i < last; i++) {
				bad_char_skip[pattern[i]] = last - i;
			}

			int startpos = ptr;
			int returnpos = startpos;

			while (hlen >= nlen) {
				// scan from the end of the needle
				for (int i = last; data[startpos + i] == pattern[i]; i--) {
					if (i == 0) { // If the first byte matches, we've found it.
						// Seek to just after the matched pattern
						ptr = startpos + nlen;
						return true;
					}
				}

				hlen -= bad_char_skip[data[startpos + last]];
				startpos += bad_char_skip[data[startpos + last]];
			}

			ptr = returnpos;
			return false;
		}
	}
}
