using System;
using System.Drawing;
using System.Drawing.Imaging;

namespace ScsMinimapper.Scs {
	/// <summary>
	/// Specialized image class to make efficient access to pixel data possible
	/// </summary>
	class RasterImage {
		private byte[] pixels;
		private int width, height;
		private int pixelSize;
		private PixelFormat format;

		/// <summary>
		/// Gets the width of the image
		/// </summary>
		public int Width {
			get { return width; }
		}

		/// <summary>
		/// Gets the height of the image
		/// </summary>
		public int Height {
			get { return height; }
		}

		/// <summary>
		/// Creates a new raster image with the specified width, height,
		/// and pixel format. For now only Format24bppRgb and
		/// Format32bppArgb are supported.
		/// </summary>
		/// <param name="width">Width of the image</param>
		/// <param name="height">Height of the image</param>
		/// <param name="format">Pixel format of the image</param>
		public RasterImage(int width, int height, PixelFormat format) {
			this.width = width;
			this.height = height;
			this.format = format;
			switch (format) {
				case PixelFormat.Format24bppRgb:
					pixelSize = 3;
					break;

				case PixelFormat.Format32bppArgb:
					pixelSize = 4;
					break;

				default:
					throw new ArgumentException(
						"Invalid format: only 24bpp rgb and 32bpp argb are supported");
			}
			pixels = new byte[width * height * pixelSize];
		}

		/// <summary>
		/// Sets the pixel at the specified coordinate to the supplied color.
		/// If the image has an alpha channel, the alpha value from the color
		/// is used. If it doesn't, it is ignored.
		/// </summary>
		/// <param name="x">X coordinate</param>
		/// <param name="y">Y coordinate</param>
		/// <param name="c">Color to use for the pixel</param>
		public void SetPixel(int x, int y, Color c) {
			int offset = (y * width + x) * pixelSize;
			pixels[offset] = c.B;
			pixels[offset + 1] = c.G;
			pixels[offset + 2] = c.R;
			if (pixelSize == 4)
				pixels[offset + 3] = c.A;
		}

		/// <summary>
		/// Sets the pixel at the specified coordinate to the supplied color.
		/// All values should be in the range 0-255.
		/// If the image has an alpha channel, the alpha value remains at
		/// the current value. Use the overload with alpha parameter instead
		/// </summary>
		/// <param name="x">X coordinate</param>
		/// <param name="y">Y coordinate</param>
		/// <param name="r">Red value</param>
		/// <param name="g">Green value</param>
		/// <param name="b">Blue value</param>
		public void SetPixel(int x, int y, int r, int g, int b) {
			int offset = (y * width + x) * pixelSize;
			pixels[offset] = (byte)b;
			pixels[offset + 1] = (byte)g;
			pixels[offset + 2] = (byte)r;
		}

		/// <summary>
		/// Sets the pixel at the specified coordinate to the supplied color.
		/// All values should be in the range 0-255.
		/// If the image doesn't have an alpha channel, the image will be
		/// corrupted. Use the overload without alpha parameter instead!
		/// </summary>
		/// <param name="x">X coordinate</param>
		/// <param name="y">Y coordinate</param>
		/// <param name="r">Red value</param>
		/// <param name="g">Green value</param>
		/// <param name="b">Blue value</param>
		/// <param name="a">Alpha value: 255 = opaque; 0 = transparent</param>
		public void SetPixel(int x, int y, int r, int g, int b, int a) {
			int offset = (y * width + x) * pixelSize;
			pixels[offset] = (byte)b;
			pixels[offset + 1] = (byte)g;
			pixels[offset + 2] = (byte)r;
			pixels[offset + 3] = (byte)a;
		}

		/// <summary>
		/// Converts the raster image to a bitmap
		/// </summary>
		/// <returns>This image converted to bitmap</returns>
		public Bitmap ToBitmap() {
			Bitmap bmp = new Bitmap(width, height, format);
			BitmapData data = bmp.LockBits(new Rectangle(0, 0, width, height),
				ImageLockMode.WriteOnly, format);
			System.Runtime.InteropServices.Marshal.Copy(pixels, 0,
				data.Scan0, pixels.Length);
			bmp.UnlockBits(data);
			return bmp;
		}
	}
}
