using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;

namespace ScsMinimapper.Scs.Readers {
	/// <summary>
	/// Class for reading the ambients and buildings from a scs file
	/// </summary>
	class ObjectReader {
		private Input input;
		private List<ScsObject> objects;
		private int size;
		private bool parsed;
		private Bitmap ambients, buildings;

		// Ambient colours
		private static Color
			a_tree = Color.FromArgb(0, 102, 0),
			a_bush = Color.FromArgb(51, 102, 51),
			a_grass = Color.FromArgb(102, 204, 51),
			a_coast = Color.FromArgb(150, 175, 130),
			a_rock = Color.FromArgb(102, 102, 102),
			a_cliff = Color.FromArgb(74, 91, 46),
			a_desert_cliff = Color.FromArgb(191, 136, 83),
			a_mesa = Color.FromArgb(153, 102, 0),
			a_ice = Color.FromArgb(102, 153, 255);

		// Building colours
		private static Color
			b_home = Color.FromArgb(255, 0, 0),
			b_workplace = Color.FromArgb(0, 255, 0),
			b_venue = Color.FromArgb(0, 0, 255),
			b_power = Color.FromArgb(255, 255, 0),
			b_decoration = Color.FromArgb(0, 255, 255),
			b_rubble = Color.FromArgb(50, 50, 50),
			b_unknown = Color.Black;

		// Blank color
		private static Color c_blank = Color.Transparent;

		/// <summary>
		/// Constructor - creates a new object reader with the specified
		/// input.
		/// </summary>
		/// <param name="input">SCS input file</param>
		/// <param name="size">Map size</param>
		public ObjectReader(Input input, int size) {
			Random random = new Random();
			this.input = input;
			this.size = size;
			objects = new List<ScsObject>();
		}

		/// <summary>
		/// Gets the ambient elements (rocks, trees) from the file
		/// and returns it as an image.
		/// </summary>
		/// <returns>Image with ambients</returns>
		public Bitmap GetAmbients() {
			if (!parsed) Parse();

			return ambients;
		}

		/// <summary>
		/// Gets the buildings from the file and returns it as an image
		/// </summary>
		/// <returns>Image with buildings</returns>
		public Bitmap GetBuildings() {
			if (!parsed) Parse();

			return buildings;
		}

		/// <summary>
		/// Parses the file, searching for objects
		/// </summary>
		private void Parse() {
			ParseObjects();
			ParseList();
			CreateBitmaps();
			parsed = true;
		}

		/// <summary>
		/// Searches the file for any objects, fills the 'objects' array
		/// </summary>
		private void ParseObjects() {
			input.Seek(0);
			byte[] pattern = { 0xef, 0xbe, 0x0d, 0x60 };
			while (input.FindPattern(pattern)) {
				ScsObject obj = new ScsObject();

				// Skip until the list of tiles
				if (input.ReadInt() == 0x811C9DC5) {
					obj.isBuilding = true;
					input.SkipBytes(8);
				}

				input.SkipBytes(4);
				obj.id = input.ReadInt();
				input.SkipBytes(1);

				int numtiles = (int)input.ReadInt();
				for (int i = 0; i < numtiles; i++) {
					int x = 1 + (int)input.ReadInt();
					int y = (int)input.ReadInt();
					obj.tiles.Add(new Point(x, y));
				}

				objects.Add(obj);
			}
		}

		/// <summary>
		/// Searches the file for the list of objects and their IDs/types.
		/// ParseObjects() must be called before this method.
		/// </summary>
		private void ParseList() {
			if (objects.Count == 0) {
				return;
			}

			input.Seek(0);
			byte[] pattern = new byte[12];
			// Pattern is 3 ints:
			// - Size of the map
			// - Number of objects
			// - ID of the first building
			IntToByteArray((uint)size, pattern, 0);
			IntToByteArray((uint)objects.Count, pattern, 4);
			IntToByteArray(objects[0].id, pattern, 8);

			if (!input.FindPattern(pattern)) {
				return;
			}
			input.SkipBytes(-4);

			try {
				for (int i = 0; i < objects.Count; i++) {
					ScsObject obj = objects[i];
					uint id = input.ReadInt();
					uint objId = input.ReadInt();
					if (i != objects.Count - 1 && objId == objects[i + 1].id) {
						// Duplication: make transparent
						obj.color = Color.Transparent;
						input.SkipBytes(-4);
						continue;
					}
					input.SkipString();
					uint typeId = input.ReadInt();
					input.SkipString();

					if (id != obj.id) {
						// Oh, we're in big trouble!
						break;
					}

					if (obj.isBuilding) {
						obj.color = LookupBuildingColor(typeId);
					} else {
						obj.color = LookupAmbientColor(objId);
					}
				}
			} catch (InputException) { }
		}

		/// <summary>
		/// Creates the ambients and buildings bitmaps from the list
		/// of objects. Requires both ParseObjects() and ParseLists()
		/// to be run before calling this method.
		/// </summary>
		private void CreateBitmaps() {
			RasterImage amb = new RasterImage(size, size, PixelFormat.Format32bppArgb);
			RasterImage bld = new RasterImage(size, size, PixelFormat.Format32bppArgb);
			foreach (ScsObject obj in objects) {
				RasterImage cur = (obj.isBuilding ? bld : amb);
				foreach (Point tile in obj.tiles) {
					cur.SetPixel(tile.X, tile.Y, obj.color);
				}
			}
			ambients = amb.ToBitmap();
			buildings = bld.ToBitmap();
		}

		/// <summary>
		/// Converts an unsigned int to a little endian byte array of length four.
		/// </summary>
		/// <param name="number">Number to convert</param>
		/// <param name="destination">Array to write the output to</param>
		/// <param name="offset">Offset in the array to start writing</param>
		private void IntToByteArray(uint number, byte[] destination, int offset) {
			destination[offset] = (byte)(number & 0xff);
			destination[offset + 1] = (byte)((number >> 8) & 0xff);
			destination[offset + 2] = (byte)((number >> 16) & 0xff);
			destination[offset + 3] = (byte)((number >> 24) & 0xff);
		}

		/// <summary>
		/// Looks up the minimap color of an ambient object.
		/// Color of ambients is determined by the ID, not the type
		/// </summary>
		/// <param name="type">ID of the ambient object</param>
		/// <returns>Minimap color of the object</returns>
		private Color LookupAmbientColor(uint type) {
			switch (type) {
				case 0x5cf3f606: return a_bush;
				case 0x6c0a2955: return a_bush;
				case 0xf82ef834: return a_bush;
				case 0xc3ee45db: return a_bush;
				case 0xd52ff3cc: return a_bush;
				case 0x1221252d: return a_bush;
				case 0x15212a6e: return a_bush;
				case 0x196410e0: return a_bush;
				case 0x57690d91: return a_bush;
				case 0x498a2195: return a_bush;
				case 0x5c8a3fb6: return a_bush;
				case 0x377679c4: return a_bush;
				case 0x4655e2b1: return a_bush;
				case 0x5835fe80: return a_bush;
				case 0x4155da9a: return a_bush;
				case 0x5327251c: return a_bush;
				case 0x5327251f: return a_bush;
				case 0x5327251e: return a_bush;
				case 0x53272519: return a_bush;
				case 0x53272518: return a_bush;
				case 0x4a7eb635: return a_coast;
				case 0x587adee4: return a_coast;
				case 0x5d7ed456: return a_coast;
				case 0xf08e559: return a_coast;
				case 0xf2032d38: return a_coast;
				case 0x1a08f6e2: return a_coast;
				case 0xf7652ef9: return a_coast;
				case 0x5b4b1458: return a_coast;
				case 0x2653f82: return a_coast;
				case 0xb0d824dd: return a_coast;
				case 0x1f44109c: return a_coast;
				case 0x93d7f77e: return a_coast;
				case 0x2ad7af9f: return a_tree;
				case 0x39d7c7f8: return a_tree;
				case 0x92c155b3: return a_bush;
				case 0x438be7b5: return a_rock;
				case 0xbb33d5c4: return a_ice;
				case 0x9497744b: return a_ice;
				case 0xb833d087: return a_ice;
				case 0xc9b32e97: return a_tree;
				case 0xc9b32e96: return a_tree;
				case 0xe2069f7d: return a_tree;
				case 0xf7945640: return a_tree;
				case 0x386d8e38: return a_tree;
				case 0x386d8e39: return a_tree;
				case 0xaeb70bda: return a_tree;
				case 0xf19aa6b: return a_rock;
				case 0x5269baea: return a_rock;
				case 0x2619ce2c: return a_rock;
				case 0xa289c93b: return a_rock;
				case 0xcf0b036d: return a_rock;
				case 0x7958d20c: return a_rock;
				case 0xd20b07ae: return a_rock;
				case 0xbea13106: return a_rock;
				case 0x2fbe0a55: return a_rock;
				case 0xa7a10d41: return a_rock;
				case 0x672c188: return a_rock;
				case 0xcca499df: return a_rock;
				case 0xfb72b0f3: return a_rock;
				case 0x924b5846: return a_rock;
				case 0x7b4b3381: return a_rock;
				case 0xfcedbb0d: return a_rock;
				case 0x553fe32c: return a_rock;
				case 0xffedc04e: return a_rock;
				case 0x468a7b3b: return a_bush;
				case 0x3d8a6d5c: return a_bush;
				case 0xd6b92436: return a_bush;
				case 0xdfb93211: return a_bush;
				case 0x789fe640: return a_bush;
				case 0x5d9fbbcb: return a_bush;
				case 0xbe3f6aa1: return a_tree;
				case 0xbe3f6aa2: return a_tree;
				case 0xbe3f6aa3: return a_tree;
				case 0x3023b442: return a_tree;
				case 0x65c90db8: return a_tree;
				case 0x65c90dbb: return a_tree;
				case 0x65c90dba: return a_tree;
				case 0xe2fd1321: return a_tree;
				case 0xe2fd1322: return a_tree;
				case 0xe2fd1323: return a_tree;
				case 0x9641bfa9: return a_tree;
				case 0x9641bfaa: return a_tree;
				case 0x9641bfab: return a_tree;
				case 0x2fc61169: return a_tree;
				case 0x2fc6116a: return a_tree;
				case 0xc39058f: return a_tree;
				case 0xc39058c: return a_tree;
				case 0xc39058d: return a_tree;
				case 0x5c4299c1: return a_tree;
				case 0x5c4299c2: return a_tree;
				case 0x5c4299c3: return a_tree;
			}
			return c_blank;
		}

		/// <summary>
		/// Looks up the minimap color of a building object.
		/// Color of buildings is determined by the type, not the ID
		/// </summary>
		/// <param name="type">Type of the building object</param>
		/// <returns>Minimap color of the object</returns>
		private Color LookupBuildingColor(uint type) {
			switch (type) {
				case 0x4a0d01a3:
					return b_decoration;
				case 0xe6bbd7de:
					return b_venue;
				case 0x8612c7f7:
					return b_workplace;
				case 0xbda06a7c:
					return b_home;
				case 0xa172d288:
					return b_power;
				case 0xaea53c4f:
					return b_rubble;
			}
			return b_unknown;
		}

		/// <summary>
		/// Private internal class for storing information about objects
		/// </summary>
		private class ScsObject {
			public uint id;
			public bool isBuilding;
			public List<Point> tiles;
			public Color color = Color.Transparent;

			public ScsObject() {
				tiles = new List<Point>();
			}
		}
	}
}
