using System;
using System.Drawing;
using System.Drawing.Imaging;
using ScsMinimapper.Scs.Terrain;

namespace ScsMinimapper.Scs.Readers {
	/// <summary>
	/// Class for reading the terrain from a scs file
	/// </summary>
	class TerrainReader {
		private Input input;

		/// <summary>
		/// Constructor - creates a new terrain reader with the specified
		/// input. The input should be set at the first byte of the
		/// terrain type grid.
		/// </summary>
		/// <param name="input">Input to read the terrain from</param>
		public TerrainReader(Input input) {
			this.input = input;
		}

		/// <summary>
		/// Reads the terrain info from the input and creates an image
		/// from it.
		/// </summary>
		/// <returns>Bitmap with terrain</returns>
		public Bitmap GetTerrain() {
			int width = (int)input.ReadInt() / 2;
			int height = (int)input.ReadInt() / 2;

			// Init random bytes
			byte[] randomness = new byte[width * height / 8];
			Random rand = new Random();
			rand.NextBytes(randomness);

			// Get right terrain
			int maxPosition = input.Length() - (72 * height * width);
			BaseTerrain terrain = GetTerrainType(maxPosition);

			RasterImage image = new RasterImage(width, height, PixelFormat.Format24bppRgb);
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					byte randomByte = randomness[(y * height + x) / 8];
					bool random = ((randomByte >> ((y * height + x) % 8)) % 2) == 1;
					image.SetPixel(x, y, terrain.GetColor(input.ReadByte(), random));
				}
			}

			return image.ToBitmap();
		}

		/// <summary>
		/// Gets the terrain type from the file
		/// </summary>
		/// <param name="maxPosition">Maximum position to seek</param>
		/// <returns>Terrain colors object</returns>
		private BaseTerrain GetTerrainType(int maxPosition) {
			// Store input state
			int position = input.Pos();

			// Immensely inefficient way to determine the terrain:
			// just try to find 'em all.

			// Temperate
			if (input.FindPattern(new byte[] { 0xb0, 0x8e, 0x46, 0xd8 }, maxPosition)) {
				input.Seek(position);
				return new TemperateTerrain();
			}

			// Temperate River
			if (input.FindPattern(new byte[] { 0x2e, 0xe7, 0x78, 0x81 }, maxPosition)) {
				input.Seek(position);
				return new TemperateRiverTerrain();
			}

			// Alpine
			if (input.FindPattern(new byte[] { 0x5e, 0x06, 0x1f, 0x82 }, maxPosition)) {
				input.Seek(position);
				return new AlpineTerrain();
			}

			// Desert
			if (input.FindPattern(new byte[] { 0xe0, 0x46, 0xf8, 0x09 }, maxPosition)) {
				input.Seek(position);
				return new DesertTerrain();
			}

			// Lush
			if (input.FindPattern(new byte[] { 0x3f, 0x01, 0x71, 0xfd }, maxPosition)) {
				input.Seek(position);
				return new LushTerrain();
			}

			// Savannah
			if (input.FindPattern(new byte[] { 0x0f, 0x01, 0xbf, 0x39 }, maxPosition)) {
				input.Seek(position);
				return new SavannahTerrain();
			}

			// Tundra
			if (input.FindPattern(new byte[] { 0x03, 0x26, 0x2b, 0x74 }, maxPosition)) {
				input.Seek(position);
				return new TundraTerrain();
			}

			// Default to Temperate
			return new TemperateTerrain();
		}
	}
}
